module beamdata
  use const_and_precisions, only : wp_
  implicit none

  integer, save :: nray,nrayr,nrayth,nstep,jray1
  real(wp_), save :: dst,h,hh,h6,rwmax,twodr2
  integer, parameter :: nfileproj0 = 8, nfilew = 12

contains

  subroutine init_rtr(rtrparam,ywork,ypwork,xc,du1,gri,ggri, &
                      psjki,tauv,alphav,ppabs,dids,ccci,p0jk,ext,eyt,iiv)
    use gray_params, only : rtrparam_type
    use const_and_precisions, only : zero,half,two
    implicit none
    type(rtrparam_type), intent(in) :: rtrparam
    real(wp_),  dimension(:,:), intent(out), allocatable :: ywork,ypwork, &
       gri,psjki,tauv,alphav,ppabs,dids,ccci
    real(wp_), dimension(:,:,:), intent(out), allocatable :: xc,du1,ggri
    real(wp_),    dimension(:), intent(out), allocatable :: p0jk
    complex(wp_), dimension(:), intent(out), allocatable :: ext, eyt
    integer,      dimension(:), intent(out), allocatable :: iiv

    dst=rtrparam%dst
    h=dst
    hh=h*half
    h6=h/6.0_wp_
    
    nrayr=rtrparam%nrayr
    nrayth=rtrparam%nrayth
    if(nrayr==1) nrayth=1
    nray=(nrayr-1)*nrayth+1
    
    rwmax=rtrparam%rwmax
    if(nrayr>1) then
      twodr2 = two*(rwmax/(nrayr-1))**2
    else
      twodr2 = two
    end if
    
    nstep=rtrparam%nstep
    
    call alloc_beam(ywork,ypwork,xc,du1,gri,ggri, &
                    psjki,tauv,alphav,ppabs,dids,ccci,p0jk,ext,eyt,iiv)
  end subroutine init_rtr

  function rayi2jk(i) result(jk)
    implicit none
    integer, intent(in) :: i
    integer, dimension(2) :: jk
    integer :: ioff

    if (i>1) then
      ioff  = i - 2
      jk(1) = ioff/nrayth              ! jr-2
      jk(2) = ioff - jk(1)*nrayth + 1  ! kt
!     jk(2) = mod(ioff,nrayth) + 1     ! kt
      jk(1) = jk(1) + 2                ! jr
    else
      jk = 1
    end if
  end function rayi2jk

  function rayi2j(i) result(jr)
    implicit none
    integer, intent(in) :: i
    integer :: jr

!   jr = max(1, (i-2)/nrayth + 2)
    if (i>1) then
      jr = (i-2)/nrayth + 2
    else
      jr = 1
    end if
  end function rayi2j

  function rayi2k(i) result(kt)
    implicit none
    integer, intent(in) :: i
    integer :: kt

!   kt = max(1, mod(i-2,nrayth) + 1)
    if (i>1) then
      kt = mod(i-2,nrayth) + 1
    else
      kt = 1
    end if
  end function rayi2k

  function rayjk2i(jr,kt) result(i)
    implicit none
    integer, intent(in) :: jr,kt
    integer :: i
    
!    i = max(1, (jr-2)*nrayth + kt + 1)
    if (jr>1) then
      i = (jr-2)*nrayth + kt + 1
    else
      i = 1
    end if
  end function rayjk2i
  
  subroutine alloc_beam(ywork,ypwork,xc,du1,gri,ggri, &
                    psjki,tauv,alphav,ppabs,dids,ccci,p0jk,ext,eyt,iiv)
    implicit none
    real(wp_),  dimension(:,:), intent(out), allocatable :: ywork,ypwork, &
       gri,psjki,tauv,alphav,ppabs,dids,ccci
    real(wp_), dimension(:,:,:), intent(out), allocatable :: xc,du1,ggri
    real(wp_),    dimension(:), intent(out), allocatable :: p0jk
    complex(wp_), dimension(:), intent(out), allocatable :: ext, eyt
    integer,      dimension(:), intent(out), allocatable :: iiv
    
    call dealloc_beam(ywork,ypwork,xc,du1,gri,ggri, &
                    psjki,tauv,alphav,ppabs,dids,ccci,p0jk,ext,eyt,iiv)

    allocate(ywork(6,nray), ypwork(6,nray), gri(3,nray), ggri(3,3,nray), &
             xc(3,nrayth,nrayr), du1(3,nrayth,nrayr),                    &
             psjki(nray,nstep),  tauv(nray,nstep),  alphav(nray,nstep),  &
             ppabs(nray,nstep),  dids(nray,nstep),    ccci(nray,nstep),  &
             p0jk(nray), ext(nray), eyt(nray), iiv(nray))
  end subroutine alloc_beam


  subroutine dealloc_beam(ywork,ypwork,xc,du1,gri,ggri, &
                    psjki,tauv,alphav,ppabs,dids,ccci,p0jk,ext,eyt,iiv)
    implicit none
    real(wp_),  dimension(:,:), intent(out), allocatable :: ywork,ypwork, &
       gri,psjki,tauv,alphav,ppabs,dids,ccci
    real(wp_), dimension(:,:,:), intent(out), allocatable :: xc,du1,ggri
    real(wp_),    dimension(:), intent(out), allocatable :: p0jk
    complex(wp_), dimension(:), intent(out), allocatable :: ext, eyt
    integer,      dimension(:), intent(out), allocatable :: iiv

    if (allocated(ywork))   deallocate(ywork)
    if (allocated(ypwork))  deallocate(ypwork)
    if (allocated(xc))      deallocate(xc)
    if (allocated(du1))     deallocate(du1)
    if (allocated(gri))     deallocate(gri)
    if (allocated(ggri))    deallocate(ggri)
    if (allocated(psjki))   deallocate(psjki)
    if (allocated(tauv))    deallocate(tauv)
    if (allocated(alphav))  deallocate(alphav)
    if (allocated(ppabs))   deallocate(ppabs)
    if (allocated(dids))    deallocate(dids)
    if (allocated(ccci))    deallocate(ccci)
    if (allocated(p0jk))    deallocate(p0jk)
    if (allocated(ext))     deallocate(ext)
    if (allocated(eyt))     deallocate(eyt)
    if (allocated(iiv))     deallocate(iiv)
  end subroutine dealloc_beam

  subroutine pweight(p0,p0jk)
! power associated to jk-th ray p0jk(j) for total beam power p0
    use const_and_precisions, only : wp_, zero, one, half, two
    implicit none
! arguments
    real(wp_), intent(in) :: p0
    real(wp_), dimension(:), intent(out) :: p0jk
! local variables
    integer :: j,jk,jkn
    real(wp_) :: dr,r,w,r0,w0,summ
    real(wp_), dimension(nrayr) :: q

    if(nray==1) then
      q(1) = one
    else
      dr = rwmax/dble(nrayr - 1)
      summ = zero
      w0 = one
      do j = 1, nrayr
        r    = (dble(j) - half)*dr
        w    = exp(-two*r**2)
        q(j) = w - w0
        summ = summ + q(j)
        r0   = r
        w0   = w
      end do
      q = q/summ
      q(2:) = q(2:)/nrayth
    end if

    p0jk(1)=q(1)*p0
    jk=2
    do j=2,nrayr
      jkn=jk+nrayth
      p0jk(jk:jkn-1)=q(j)*p0
      jk=jkn
    end do
  end subroutine pweight

  subroutine print_projxyzt(st,ywrk,iproj)
    use const_and_precisions, only : wp_, comp_huge, zero, one
    implicit none
! arguments
    real(wp_), intent(in) :: st
    real(wp_), dimension(:,:), intent(in) :: ywrk
    integer, intent(in) :: iproj
! local variables
    integer :: jk,jkz,nfile
    integer, dimension(2) ::jkv
    real(wp_), dimension(3) :: xv1,dir,dxv
    real(wp_) :: dirm,rtimn,rtimx,csth1,snth1,csps1,snps1,xti,yti,zti,rti
! common/external functions/variables

    nfile = nfileproj0 + iproj

    xv1  = ywrk(1:3,1)
    dir  = ywrk(4:6,1)
    dirm = sqrt(dir(1)**2 + dir(2)**2 + dir(3)**2)
    dir  = dir/dirm
    csth1 = dir(3)
    snth1 = sqrt(one - csth1**2)
    if(snth1 > zero) then
      csps1=dir(2)/snth1
      snps1=dir(1)/snth1
    else
      csps1=one
      snps1=zero
    end if

    if(iproj==0) then
      jkz = nray - nrayth + 1
    else
      jkz = 1
    end if

    rtimn = comp_huge
    rtimx = zero
    do jk = jkz, nray
      dxv = ywrk(1:3,jk) - xv1
      xti = dxv(1)*csps1 - dxv(2)*snps1
      yti =(dxv(1)*snps1 + dxv(2)*csps1)*csth1 - dxv(3)*snth1
      zti =(dxv(1)*snps1 + dxv(2)*csps1)*snth1 + dxv(3)*csth1
      rti = sqrt(xti**2 + yti**2)

      jkv=rayi2jk(jk)
      if(.not.(iproj==0 .and. jk==1))                  &
         write(nfile,'(1x,e16.8e3,2i5,4(1x,e16.8e3))') st,jkv,xti,yti,zti,rti
      if(iproj==1 .and. jkv(2)==nrayth) write(nfile,*) ' '

      if(rti>=rtimx .and. jkv(1)==nrayr) rtimx = rti
      if(rti<=rtimn .and. jkv(1)==nrayr) rtimn = rti
    end do
    write(nfile,*) ' '
    write(nfilew,'(3(1x,e16.8e3))') st,rtimn,rtimx
  end subroutine print_projxyzt

end module beamdata
